/*
 *******************************************************************************
 *                                                                             *
 *   configuration information:  the following variables need to be set to     *
 *   indicate the machine configuration/capabilities.                          *
 *                                                                             *
 *   HAS_64BITS:  define this for a machine that has true 64-bit hardware      *
 *   including leading-zero hardware, population count, etc.  ie, a Cray-like  *
 *   machine.                                                                  *
 *                                                                             *
 *   UNIX:  define this if the program is being run on a unix-based system,    *
 *   which causes the executable to use unix-specific runtime utilities.       *
 *                                                                             *
 *   CPUS=N:  this sets up data structures to the appropriate size to support  *
 *   up to N simultaneous search engines.  note that you can set this to a     *
 *   value larger than the max processors you currently have, because the mt=n *
 *   command (added to the command line or your crafty.rc/.craftyrc file) will *
 *   control how many threads are actually spawned.                            *
 *                                                                             *
 *******************************************************************************
 */
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#if !defined(TYPES_INCLUDED)
#  include "lock.h"
#  if defined (_MSC_VER) && (_MSC_VER >= 1300) && \
    (!defined(_M_IX86) || (_MSC_VER >= 1400))
#    define RESTRICT __restrict
#  else
#    define RESTRICT
#  endif
#  if !defined(CPUS)
#    define CPUS 1
#  endif
#  if defined(WINDOWS)
#    include <windows.h>
#    include <process.h>
#  endif
#  define TYPES_INCLUDED
#  define CDECL
#  define STDCALL
/* Provide reasonable defaults for UNIX systems. */
#  undef  HAS_64BITS    /* machine has 64-bit integers / operators    */
#  define UNIX  /* system is unix-based                       */
/* Architecture-specific definitions */
#  if defined(AIX)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    define UNIX        /* system is unix-based                       */
#  endif
#  if defined(AMIGA)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    undef  UNIX        /* system is unix-based                       */
#  endif
#  if defined(FreeBSD)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    define UNIX        /* system is unix-based                       */
#  endif
#  if defined(HP)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    define UNIX        /* system is unix-based                       */
#  endif
#  if defined(LINUX)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    define UNIX        /* system is unix-based                       */
#  endif
#  if defined(MIPS)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    define UNIX        /* system is unix-based                       */
#  endif
#  if defined(NetBSD)
#    if defined(__alpha__)
#      define HAS_64BITS        /* machine has 64-bit integers / operators   */
#      define UNIX      /* system is unix-based                      */
#    else
#      undef  HAS_64BITS        /* machine has 64-bit integers / operators   */
#      define UNIX      /* system is unix-based                      */
#    endif
#  endif
#  if defined(NEXT)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    define UNIX        /* system is unix-based                       */
#  endif
#  if defined(WINDOWS)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    undef  UNIX        /* system is unix-based                       */
#    undef  STDCALL
#    define STDCALL __stdcall
#    ifdef  VC_INLINE
#      undef  CDECL
#      define CDECL __cdecl
#    endif
#  endif
#  if defined(OS2)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    define UNIX        /* system is unix-based                       */
#  endif
#  if defined(SGI)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    define UNIX        /* system is unix-based                       */
#  endif
#  if defined(SUN)
#    undef  HAS_64BITS  /* machine has 64-bit integers / operators    */
#    define UNIX        /* system is unix-based                       */
#  endif
#  if !defined(BOOKDIR)
#    define     BOOKDIR        "."
#  endif
#  if !defined(LOGDIR)
#    define      LOGDIR        "./log"
#  endif
#  if !defined(TBDIR)
#    define       TBDIR     "./TB"
#  endif
#  if !defined(RCDIR)
#    define       RCDIR        "."
#  endif
#  if !defined(NOEGTB)
#    define     EGTB_CACHE_DEFAULT               1024*1024
#  endif
#  define     MAXPLY                                  260
#  define     MAXPLY_ULLAGE                           140
#  define     MAX_TC_NODES                        3000000
#  define     MAX_BLOCKS_PER_CPU                      128
#  define     MAX_BLOCKS         (MAX_BLOCKS_PER_CPU*CPUS)
#  define      BOOK_CLUSTER_SIZE           8000
#  define     BOOK_POSITION_SIZE             16
#  define            MERGE_BLOCK           1000
#  define             SORT_BLOCK        4000000
#  define         LEARN_INTERVAL             10
#  define      LEARN_COUNTER_BAD            -80
#  define     LEARN_COUNTER_GOOD           +100
#  define BIG_SCORE                      999999
#  define NO_VALUE                        40001
#  define INFTY                           32769
#  define MATE                            32768
#  define NOT_MATE                        32468
#  define PAWN_VALUE                        100
#  define KNIGHT_VALUE                      325
#  define BISHOP_VALUE                      325
#  define ROOK_VALUE                        500
#  define QUEEN_VALUE                       975
#  define KING_VALUE                      40000
#  define IMBALANCE_MINOR                    88 // Minor down
#  define IMBALANCE_MAJOR                   104 // Major down
#  define IMBALANCE_MINOR_MAJOR              42 // Minor -v- Major (and other small imbalances)
#  define IMBALANCE_MAX_SCORE               126 // Maximum score
#  define BISHOP_PAIR_MG                     30
#  define BISHOP_PAIR_EG                     30
#  define ASPIRATION_WINDOW                  40
#  define EG_MAT                             13
#  define TOT_MAT                            39
#  define MAX_DRAFT                         256
#  define TIMER_SEARCH_CENTISECONDS          10
#  define TIMER_TRACE_CENTISECONDS         6000
#  define HISTORY_MAX_MOVES                  10
#  define ITERATE_FAIL_DELTA                  8
#  define NNUE_ARRAY                         33
#  if defined(HAS_64BITS)
typedef unsigned long BITBOARD;
typedef unsigned long UINT64;
#  elif defined(WINDOWS)
typedef unsigned __int64 BITBOARD;
typedef unsigned __int64 UINT64;
#  else
typedef unsigned long long BITBOARD;
typedef unsigned long long UINT64;
#  endif
#  if defined(WINDOWS)
#    define BMF   "%I64u"
#    define BMF6  "%6I64u"
#    define BMF10 "%10I64u"
#  else
#    define BMF   "%llu"
#    define BMF6  "%6llu"
#    define BMF10 "%10llu"
#  endif
#  if defined(UNIX) & (CPUS > 1)
#    include <pthread.h>
#  endif
#  include <time.h>
#  if !defined(CLOCKS_PER_SEC)
#    define CLOCKS_PER_SEC 1000000
#  endif
typedef enum {
  A1, B1, C1, D1, E1, F1, G1, H1,
  A2, B2, C2, D2, E2, F2, G2, H2,
  A3, B3, C3, D3, E3, F3, G3, H3,
  A4, B4, C4, D4, E4, F4, G4, H4,
  A5, B5, C5, D5, E5, F5, G5, H5,
  A6, B6, C6, D6, E6, F6, G6, H6,
  A7, B7, C7, D7, E7, F7, G7, H7,
  A8, B8, C8, D8, E8, F8, G8, H8,
  BAD_SQUARE
} squares;
typedef enum { FILEA, FILEB, FILEC, FILED, FILEE, FILEF, FILEG, FILEH } files;
typedef enum { RANK1, RANK2, RANK3, RANK4, RANK5, RANK6, RANK7, RANK8 } ranks;
typedef enum { empty = 0, occupied = 0, pawn = 1, knight = 2, bishop = 3,
  rook = 4, queen = 5, king = 6
} PIECE;
typedef enum { nnue_empty = 0, nnue_wking = 1, nnue_wqueen = 2, nnue_wrook = 3,
  nnue_wbishop = 4, nnue_wknight = 5, nnue_wpawn = 6, nnue_bking = 7, nnue_bqueen = 8,
  nnue_brook = 9, nnue_bbishop = 10, nnue_bknight = 11, nnue_bpawn = 12
} NNUE_PIECE;
typedef enum { black = 0, white = 1 } COLOR;
typedef enum { nnue_white = 0, nnue_black = 1 } NNUE_COLOR;
typedef enum { mg = 0, eg = 1 } PHASE;
typedef enum { lb = 0, ub = 1 } BOUND;
typedef enum { empty_v = 0, pawn_v = 1, knight_v = 3,
  bishop_v = 3, rook_v = 5, queen_v = 9, king_v = 99
} PIECE_V;
typedef enum { think = 1, puzzle = 2, book = 3, annotate = 4 } SEARCH_TYPE;
typedef enum { normal_mode, tournament_mode } PLAYING_MODE;
typedef enum { book_learning = 1, result_learning = 2
} LEARNING_MODE;
typedef enum { PV_NODE = 0, CUT_NODE = 1, ALL_NODE = -1} NODE_TYPE;
typedef struct {
  unsigned char enpassant_target;
  signed char castle[2];
  unsigned char rule_50_moves;
} SEARCH_POSITION;
typedef struct {
  int move1;
  int move2;
} KILLER;
typedef struct {
  BITBOARD pieces[7];
} BB_PIECES;
typedef struct {
  BB_PIECES color[2];
  BITBOARD rooks_queens;
  BITBOARD bishops_queens;
  UINT64 hash_key;
  UINT64 pawn_hash_key;
  int material_evaluation;
  int kingsq[2];
  signed char board[64];
  signed char pieces[2][7];
  signed char pawns[2];
  signed char total_all_pieces;
} POSITION;
typedef struct {
  UINT64 word1;
  UINT64 word2;
} HASH_ENTRY;
typedef struct {
  UINT64 key;
  int score_mg, score_eg;
  unsigned char defects_k[2];
  unsigned char defects_e[2];
  unsigned char defects_d[2];
  unsigned char defects_q[2];
  unsigned char all[2];
  unsigned char passed[2];
  unsigned char candidates[2];
  unsigned char open_files;
  unsigned char filler;
} PAWN_HASH_ENTRY;
typedef struct {
  UINT64 entry[4];
} PXOR;
typedef struct {
  int path[MAXPLY];
  unsigned char pathh;
  unsigned char pathl;
  unsigned char pathd;
} PATH;
typedef struct {
  UINT64 path_sig;
  int hash_pathl;
  int  hash_path_age;
  int hash_path_moves[MAXPLY];
} HPATH_ENTRY;
typedef struct {
  int phase;
  int captures;
  int remaining;
  int *last;
  int done[13]; // 1 hash, 4 mate killer moves, 4 killer moves 2 counter moves and 2 pair moves
  int excluded;
  int history;
} NEXT_MOVE;
typedef struct {
  UINT64 nodes;
  int move;
/*
    1 x..xx xxxx xxx1 = failed low this iteration
    2 x..xx xxxx xx1x = failed high this iteration
    4 x..xx xxxx x1xx = don't search in parallel
    8 x..xx xxxx 1xxx = do not reduce this move
   16 x..xx xxx1 xxxx = move has been searched
 */
  unsigned int status;
  int bm_age;
} ROOT_MOVE;

#  if defined(WINDOWS)
#    pragma pack(4)
#  endif
typedef struct {
  BITBOARD position;
  unsigned int status_played;
  float learn;
} BOOK_POSITION;

#  if defined(WINDOWS)
#    pragma pack()
#  endif
typedef struct {
  unsigned char position[8];
  unsigned char status;
  unsigned char percent_play;
} BB_POSITION;
struct personality_term {
  char *description;
  int type;
  int size;
  int *value;
  int *value_last;
  int category;
};
typedef struct {
  int move;
  int value;
  int type;
  int draft;
} HASH;
struct tree {
  POSITION pos;
  UINT64 save_hash_key[MAXPLY + 3];
  BITBOARD rep_list[2][128];
  BITBOARD all_pawns;
  UINT64 nodes_searched;
  UINT64 quiescent_nodes_searched;
  UINT64 save_pawn_hash_key[MAXPLY + 3];
  BITBOARD cache_n[64];
  BITBOARD cache_b_friendly[64];
  BITBOARD cache_b_enemy[64];
  BITBOARD cache_r_friendly[64];
  BITBOARD cache_r_enemy[64];
  BITBOARD cache_q[64];
  PAWN_HASH_ENTRY pawn_score;
  SEARCH_POSITION position[MAXPLY + 3];
  NEXT_MOVE next_status[MAXPLY];
  PATH pv[MAXPLY];
  int cache_n_mobility[64];
  int cache_b_mobility[64];
  int cache_r_mobility_mg[64];
  int cache_r_mobility_eg[64];
  int cache_q_mobility[64];
  int rep_index[2];
  int curmv[MAXPLY];
  int hash_move[MAXPLY];
  int *last[MAXPLY];
  unsigned int fail_high;
  unsigned int fail_high_first;
  unsigned int evaluations;
  unsigned int egtb_probes;
  unsigned int egtb_probes_successful;
  unsigned int extensions_done;
  unsigned int qchecks_done;
  unsigned int reductions_done;
  KILLER killers[MAXPLY];
  int move_list[16384];
  int sort_value[256];
  signed char inchk[MAXPLY];
  signed char phase[MAXPLY];
  int search_value;
  int tropism[2];
  int dangerous[2];
  int score_mg, score_eg;
  int root_move;
#  if (CPUS > 1)
  lock_t lock;
#  endif
  long thread_id;
  volatile int stop;
  char root_move_text[16];
  char remaining_moves_text[16];
  struct tree *volatile siblings[CPUS], *parent;
  volatile int nprocs;
  int alpha;
  int beta;
  int value;
  int wtm;
  int depth;
  int ply;
  int cutmove;
  int *searched;
  int moves_searched;
  int best_capture;
  NODE_TYPE node_type;
  volatile int used;
};
typedef struct tree TREE;
#if defined(_WIN32) || defined(_WIN64)
struct _clock {
  int check[CPUS];
  UINT64 search_check[CPUS];
  UINT64 time_check[CPUS];
};
typedef struct _clock CLOCK;
#endif
/*
   DO NOT modify these.  these are constants, used in multiple modules.
   modification may corrupt the search in any number of ways, all bad.
 */
#  define WORTHLESS                 0
#  define UNKNOWN                   0
#  define LOWER                     1
#  define UPPER                     2
#  define EXACT                     3
#  define HASH_MISS                 0
#  define HASH_HIT                  1
#  define AVOID_NULL_MOVE           2
#  define DO_NULL                   1
#  define NO_NULL                   0
#  define DO_NNUE                   1
#  define NO_NNUE                   0
#  define DO_SINGULAR               1
#  define NO_SINGULAR               0
#  define DO_DELTA                  1
#  define NO_DELTA                  0
#  define NO_CHECK                  0
#  define NONE                      0
#  define NULL_MOVE                 0
#  define EVALUATION                0
#  define NO_MOVE                   0
#  define VOID_MOVE                 1
#  define HASH_MOVE                 2
#  define GENERATE_CAPTURE_MOVES    3
#  define CAPTURE_MOVES             4
#  define KILLER_MOVE_1             5
#  define KILLER_MOVE_2             6
#  define KILLER_MOVE_3             7
#  define KILLER_MOVE_4             8
#  define COUNTER_MOVE1             9
#  define COUNTER_MOVE2            10
#  define MOVE_PAIR1               11
#  define MOVE_PAIR2               12
#  define BAD_CAPTURE_MOVES        13
#  define GENERATE_QUIET_MOVES     14
#  define HISTORY_MOVES            15
#  define REMAINING_MOVES          16
#  define LATE_MOVES               17
#  define DO_NOTHING                0
/*
   pruning search flags
 */
#  define DO_NULL_MOVE              1    /* do null-move pruning                          */
#  define DO_FUTILITY               2    /* do futility pruning                           */
#  define DO_REDUCTIONS             4    /* do LMR                                        */
#  define DO_PROBCUT                8    /* do ProbCut pruning                            */
/*
   quiescent search flags
 */
#  define DO_QCHECKS                1    /* do quiescent checks                           */
/*
   search flags
 */
#  define DO_INTERNAL_DEEPEN        1    /* do internal iterative deepening               */
/*
   parameter category flags
 */
#  define CATEGORY_SEARCH           1    /* :s 's - s.a s.e s.n s.f s.r s.q               */
#  define CATEGORY_MATERIAL         2    /* :m 'm - m.P m.N m.B m.R m.Q m.i m.e           */
#  define CATEGORY_OTHER            4    /* :o 'o - o.w o.d o.l o.e                       */
#  define CATEGORY_PAWN             8    /* :P 'P -                                       */
#  define CATEGORY_KNIGHT          16    /* :N 'N -                                       */
#  define CATEGORY_BISHOP          32    /* :B 'B -                                       */
#  define CATEGORY_ROOK            64    /* :R 'R -                                       */
#  define CATEGORY_QUEEN          128    /* :Q 'Q -                                       */
#  define CATEGORY_KING           256    /* :K 'K -                                       */
#  define CATEGORY_DEVELOPMENT    512    /* :d 'd -                                       */
#  define CATEGORY_SQUARE        1024    /* :t 't -                                       */
#  define CATEGORY_MOBILITY      2048    /* :a 'a -                                       */
#  define CATEGORY_PASSED_PAWN   4096    /* :p 'p -                                       */
#  define CATEGORY_TACTICS       8192    /* :c 'c -                                       */
#  define CATEGORY_STRATEGY     16384    /* :i 'i -                                       */
/*
   search extension depth indices
 */
#  define EXTENSION_CHECK             0    /* check extension                               */
#  define EXTENSION_ONE_EVASION       1    /* one check evasion extension                   */
#  define EXTENSION_RECAPTURE         2    /* recapture extension                           */
#  define EXTENSION_PAWN_RECAPTURE    3    /* pawn recapture extension                      */
#  define EXTENSION_CAPTURE_ENDGAME   4    /* capture to endgame extension                  */
#  define EXTENSION_PASSED_PAWN       5    /* passed pawn extension                         */
#  define EXTENSION_KICK              6    /* kick extension                                */
#  define EXTENSION_CASTLING          8    /* castling extension                            */
/*
   pruning proportions indices
 */
#  define PRUNING_PROPORTION_LMR_DEPTH_BIAS                  0    /* LMR_depth_bias                 */
#  define PRUNING_PROPORTION_LMR_MOVES_BIAS                  1    /* LMR_moves_bias                 */
#  define PRUNING_PROPORTION_LMR_SCALE                       2    /* LMR_scale                      */
#  define PRUNING_PROPORTION_LMR_PV_SCALE                    3    /* LMR_PV_scale                   */
#  define PRUNING_PROPORTION_LMR_LATE_MOVES_SCALE            4    /* LMR_late_moves_scale           */
#  define PRUNING_PROPORTION_LMR_LATE_MOVES_REDUCTION_SCALE  5    /* LMR_late_moves_reduction_scale */
/*
   etc.
*/
#  define CAN_WIN_WHITE             1
#  define CAN_WIN_BLACK             2
#  define CANNOT_WIN_BLACK          1
#  define CANNOT_WIN_WHITE          2
#  define CAN_WIN (CAN_WIN_WHITE | CAN_WIN_BLACK)
#  define CANNOT_WIN                0
#  if defined(VC_INLINE)
#    include "vcinline.h"
#  else
#    if !defined(INLINE64) && !defined(INLINE32)
int CDECL PopCnt(BITBOARD);
int CDECL MSB(BITBOARD);
int CDECL LSB(BITBOARD);
#    endif
#  endif
int AbsolutePin(TREE * RESTRICT, int, int, int *);
void AlignedMalloc(void **, int, size_t);
void AlignedRemalloc(void **, int, size_t);
void Analyze(void);
void Annotate(void);
void AnnotateHeaderHTML(char *, FILE *);
void AnnotateFooterHTML(FILE *);
void AnnotatePositionHTML(TREE * RESTRICT, int, FILE *);
char *AnnotateVtoNAG(int, int, int, int);
void AnnotateHeaderTeX(char *, FILE *);
void AnnotateFooterTeX(FILE *);
void AnnotatePositionTeX(TREE *, int, FILE *);
UINT64 atoiKM(char *);
int Attacks(TREE * RESTRICT, int, int);
int DoubleCheck(TREE * RESTRICT, int);
int AttacksSliding(TREE * RESTRICT, int, int);
BITBOARD AttacksTo(TREE * RESTRICT, int);
int Average(int, int);
int *AverageVector(int *, int *, int);
void Bench(void);
int Book(TREE * RESTRICT, int, int);
void BookClusterIn(FILE *, int, BOOK_POSITION *);
void BookClusterOut(FILE *, int, BOOK_POSITION *);
int BookIn32(unsigned char *ch);
float BookIn32f(unsigned char *ch);
UINT64 BookIn64(unsigned char *ch);
int BookMask(char *);
unsigned char *BookOut32(int val);
unsigned char *BookOut32f(float val);
unsigned char *BookOut64(UINT64 val);
int BookPonderMove(TREE * RESTRICT, int);
void BookUp(TREE * RESTRICT, int, char **);
void BookSort(BB_POSITION *, int, int);
#  if defined(WINDOWS)
int _cdecl BookUpCompare(const void *, const void *);
#  else
int BookUpCompare(const void *, const void *);
#  endif
BB_POSITION BookUpNextPosition(int, int);
int CheckInput(void);
void ClearHashTableScores(void);
void ClearHashTable(void);
void CopyFromChild(TREE * RESTRICT, TREE * RESTRICT, int);
TREE *CopyToChild(TREE * RESTRICT, int);
void CraftyExit(int);
int *CrossoverVector(int *, int *, int, int);
int CrudeScore(TREE * RESTRICT, int, int, int);
void DisplayArray(int *, int);
void DisplayArrayX2(int *, int *, int);
void DisplayBitBoard(BITBOARD);
void Display2BitBoards(BITBOARD, BITBOARD);
void DisplayChessBoard(FILE *, POSITION);
void DisplayPositionFEN(FILE *, TREE * RESTRICT, int, int);
char *DisplayEvaluation(int, int);
char *DisplayEvaluationKibitz(int, int);
void DisplayFT(int, int, int);
char *DisplayHHMM(unsigned int);
char *DisplayHHMMSS(unsigned int);
char *DisplayKM(unsigned int);
void DisplayPV(TREE * RESTRICT, int, int, unsigned int, int, PATH *);
char *DisplayTime(unsigned int);
char *DisplayTimeKibitz(unsigned int);
void DisplayTreeState(TREE * RESTRICT, int, int, int);
void DisplayChessMove(char *, int);
char *Display2Times(unsigned int);
int Drawn(TREE * RESTRICT, int);
int DoRandomly(int);
int DoCrossover(int, int);
void DisplayType3(int *, int *);
void DisplayType4(int *, int *);
void DisplayType5(int *, int *, int);
void DisplayType6(int *, int *);
void DisplayType7(int *, int *);
void DisplayType8(int *, int);
void DisplayType9(int *, int);
void DisplayVector(int *, int);
void Edit(void);
#  if !defined(NOEGTB)
int EGTBProbe(TREE * RESTRICT, int, int, int *);
void EGTBPV(TREE * RESTRICT, int);
#  endif
int Evaluate(TREE * RESTRICT, int, int, int, int);
void EvaluateBishops(TREE * RESTRICT, int);
void EvaluateDevelopment(TREE * RESTRICT, int, int);
int EvaluateDraws(TREE * RESTRICT, int, int, int, int);
void EvaluateKings(TREE * RESTRICT, int, int);
int EvaluateKingsFile(TREE * RESTRICT, int, int);
void EvaluateKnights(TREE * RESTRICT, int);
void EvaluateMate(TREE * RESTRICT, int);
void EvaluateMaterial(TREE * RESTRICT);
void EvaluatePassedPawns(TREE * RESTRICT, int);
void EvaluatePassedPawnRaces(TREE * RESTRICT, int);
void EvaluatePawns(TREE * RESTRICT, int);
void EvaluateQueens(TREE * RESTRICT, int);
void EvaluateRooks(TREE * RESTRICT, int);
void EvaluateSpace(TREE * RESTRICT, int, int);
int EvaluateWinningChances(TREE * RESTRICT, int, int);
int EvaluateNNUE(TREE * RESTRICT, int, int, int, int);
void EVTest(char *);
int Exclude(TREE *RESTRICT, int, int);
int FindBlockID(TREE * RESTRICT);
char *FormatPV(TREE * RESTRICT, int, PATH);
int FTbSetCacheSize(void *, unsigned long);
int GameOver(int);
int Gain(TREE * RESTRICT, int, int, int);
int *GenerateCaptures(TREE * RESTRICT, int, int, int *);
int *GeneratePawnNoncaptures(TREE * RESTRICT, int, int, int *);
int *GenerateCheckEvasions(TREE * RESTRICT, int, int, int *);
int *GenerateNoncaptureChecks(TREE * RESTRICT, int, int, int *);
int *GenerateNoncaptureUnderpromotions(TREE * RESTRICT, int, int, int *);
int *GenerateNoncaptures(TREE * RESTRICT, int, int, int *);
int HashProbe(TREE * RESTRICT, int, int, int, int, int, HASH *);
void HashStore(TREE * RESTRICT, int, int, int, int, int, int);
void HashStorePV(TREE * RESTRICT, int, int);
void History(TREE *RESTRICT, int, int, int, int, int*, int);
void HistoryAge(void);
int EvaluateHasOpposition(int, int, int);
int IInitializeTb(char *);
int Inherit(int, int);
int *InheritVector(int *, int *, int);
void Initialize(void);
void InitializeAttackBoards(void);
void InitializeChessBoard(TREE *);
int InitializeFindAttacks(int, int, int);
int InitializeGetLogID();
void InitializeHashTables(void);
void InitializeHistory(void);
void InitializeKillers(void);
void InitializeKingSafety(void);
void InitializeLMR(void);
void InitializePopCnt(void);
void InitializePieceValues(void);
void InitializeIterateFailDelta();
void InitializeMVVLVA(void);
void InitializePhase(void);
void InitializeRepetition(void);
void InitializeMaterialImbalance(void);
void InitializeMagic(void);
BITBOARD InitializeMagicBishop(int, BITBOARD);
BITBOARD InitializeMagicRook(int, BITBOARD);
BITBOARD InitializeMagicOccupied(int *, int, BITBOARD);
void InitializeMasks(void);
void InitializePawnMasks(void);
void InitializeRandomHash(void);
void InitializeSMP(void);
int InputMove(TREE * RESTRICT, char *, int, int, int, int);
int InputMoveICS(TREE * RESTRICT, char *, int, int, int, int);
BITBOARD InterposeSquares(int, int, int);
void Interrupt(int);
int InvalidPosition(TREE * RESTRICT);
int IsActive(PIECE piece, BITBOARD moves);
int Iterate(int, int, int);
void Kibitz(int, int, int, int, int, UINT64, int, char *);
void Killer(TREE * RESTRICT, int, int);
int KingPawnSquare(int, int, int, int);
void LearnBook(int, int, int, int);
int LearnFunction(int, int, int, int);
int LessCrudeScore(TREE * RESTRICT, int, int, int, int, int);
void MakeMove(TREE * RESTRICT, int, int, int);
void MakeMoveRoot(TREE * RESTRICT, int, int);
int Morph(int, int);
int *MorphVector(int *, int, int, int);
int Move(TREE * RESTRICT, int, int, int, int, int, NODE_TYPE, int);
int Mutate(int, int);
int *MutateVector(int *, int, int, int);
void NewGame(int);
int NextEvasion(TREE * RESTRICT, int, int);
int NextMove(TREE * RESTRICT, int, int, int, int, int, NODE_TYPE, int *);
int NextRootMove(TREE * RESTRICT, TREE * RESTRICT, int, int *);
int NextRootMoveParallel(void);
int NNUEEvaluate(TREE * RESTRICT, int, int, int, int);
int Option(TREE * RESTRICT);
int OptionMatch(char *, char *);
void OptionPerft(TREE * RESTRICT, int, int, int);
void Output(TREE * RESTRICT, int, int);
char *OutputMove(TREE * RESTRICT, int, int, int);
char *OutputMoveICS(int);
int OutputGood(TREE * RESTRICT, char *, int, int);
int ParseTime(char *);
void Pass(void);
int PinnedOnKing(TREE * RESTRICT, int, int);
int Ponder(int);
void PreEvaluate(TREE * RESTRICT);
void Print(int, char *, ...);
char *PrintKM(size_t, int);
int Quiesce(TREE * RESTRICT, int, int, int, int, int, int);
unsigned int Random32(void);
UINT64 Random64(void);
void Randomize(void);
int Read(int, char *);
int ReadChessMove(TREE * RESTRICT, FILE *, int, int);
void ReadClear(void);
unsigned int ReadClock(void);
#if defined(WINDOWS)
unsigned int ReadCPUClock(void);
#endif
int ReadPGN(FILE *, int);
int ReadNextMove(TREE * RESTRICT, char *, int, int);
int ReadParse(char *, char *args[], char *);
#if defined(_WIN32) || defined(_WIN64)
int ReadParseWindowsPath(char *, char *args[]);
#endif
int ReadInput(void);
int RepetitionCheck(TREE * RESTRICT, int, int);
int RepetitionCheckBook(TREE * RESTRICT, int, int);
int RepetitionDraw(TREE * RESTRICT, int, int);
void ResignOrDraw(TREE * RESTRICT, int);
void RestoreGame(void);
void RootMoveList(int);
int ScaleSV(int);
int Search(TREE * RESTRICT, int, int, int, int, int, int);
int SearchExtension(TREE * RESTRICT, int, int, int, int, NODE_TYPE, HASH *);
int SearchReduction(TREE * RESTRICT, int, int, int, int, int, int, NODE_TYPE);
int SearchRoot(TREE * RESTRICT, int, int, int, int);
int SearchParallel(TREE * RESTRICT, int, int, int, int, int, int);
void Trace(TREE * RESTRICT, int, int, int, int, int, char *, int);
int SetBoard(TREE *, int, char **, int);
void SetChessBitBoards(TREE *);
int SetPersonalityCategoryFlags(char *);
int SetRootAlpha(int, int);
int SetRootBeta(int, int);
void SharedFree(void *address);
int StrCnt(char *, char);
int Swap(TREE * RESTRICT, int, int);
int SwapO(TREE * RESTRICT, int, int);
void Test(char *);
void TestEPD(char *);
int Thread(TREE * RESTRICT);
void WaitForAllThreadsInitialized(void);
void *STDCALL ThreadInit(void *);
#  if defined(_WIN32) || defined(_WIN64)
void ThreadMalloc(int);
#  endif
void ThreadStop(TREE * RESTRICT);
int ThreadWait(long, TREE * RESTRICT);
void TimeAdjust(int, int);
int TimeCheck(TREE * RESTRICT, int);
void TimeSet(TREE * RESTRICT, int);
#if defined(_WIN32) || defined(_WIN64)
void TimeProcSearch(PVOID data, BOOLEAN flag);
void TimeProcTrace(PVOID data, BOOLEAN flag);
#endif
void UnmakeMove(TREE * RESTRICT, int, int, int);
int ValidMove(TREE * RESTRICT, int, int, int);
int VerifyMove(TREE * RESTRICT, int, int, int);
void ValidatePosition(TREE * RESTRICT, int, int, char *);
#  if defined(_WIN32) || defined(_WIN64)
extern void *WinMallocInterleaved(size_t, int);
extern void WinFreeInterleaved(void *, size_t);
#    define MallocInterleaved(cBytes, cThreads)\
    WinMallocInterleaved(cBytes, cThreads)
#    define FreeInterleaved(pMemory, cBytes)\
    WinFreeInterleaved(pMemory, cBytes)
#  else
#    if defined(NUMA)
#      define MallocInterleaved(cBytes, cThreads) numa_alloc_interleaved(cBytes)
#      define FreeInterleaved(pMemory, cBytes)    numa_free(pMemory, 1)
#    else
#      define MallocInterleaved(cBytes, cThreads) malloc(cBytes)
#      define FreeInterleaved(pMemory, cBytes)    free(pMemory)
#    endif
#  endif
int Zugzwang(TREE * RESTRICT tree, int wtm);
#  define Abs(a)    (((a) > 0) ? (a) : -(a))
#  define Max(a,b)  (((a) > (b)) ? (a) : (b))
#  define Min(a,b)  (((a) < (b)) ? (a) : (b))
#  define FileDistance(a,b) Abs(File(a) - File(b))
#  define RankDistance(a,b) Abs(Rank(a) - Rank(b))
#  define Distance(a,b) Max(FileDistance(a,b), RankDistance(a,b))
#  define DrawScore(wtm)                 (draw_score[wtm])
#  define PopCnt8Bit(a) (pop_cnt_8bit[a])
#  define MSB8Bit(a) (msb_8bit[a])
#  define LSB8Bit(a) (lsb_8bit[a])
#  define HistoryIndex(m) ((Piece(m) << 6) + To(m))
/*
  side = side to move
  mptr = pointer into move list
  m = bit vector of to squares to unpack
  t = pre-computed from + moving piece
 */
#  define Unpack(side, mptr, m, t)                                         \
  for ( ; m ; Clear(to, m)) {                                              \
    to = Advanced(side, m);                                                \
    *mptr++ = t | (to << 6) | (Abs(PcOnSq(to)) << 15);                     \
  }
#  define Check(wtm) Attacks(tree, KingSQ(wtm), Flip(wtm))
#  define Attack(from,to) (!(obstructed[from][to] & OccupiedSquares))
#  define AttacksBishop(square, occ) *(magic_bishop_indices[square]+((((occ)&magic_bishop_mask[square])*magic_bishop[square])>>magic_bishop_shift[square]))
#  define AttacksKnight(square) knight_attacks[square]
#  define AttacksQueen(square, occ)   (AttacksBishop(square, occ)|AttacksRook(square, occ))
#  define AttacksRook(square, occ) *(magic_rook_indices[square]+((((occ)&magic_rook_mask[square])*magic_rook[square])>>magic_rook_shift[square]))
#  define Rank(x)                 ((x)>>3)
#  define File(x)                 ((x)&7)
#  define Flip(x)                 ((x)^1)
#  define PawnAttacks(wtm, x)     (pawn_attacks[Flip(wtm)][(x)] & Pawns(wtm))
#  define Advanced(wtm, pawns)    ((wtm) ? MSB(pawns) : LSB(pawns))
#  define MostAdvanced(wtm, sqs)  ((wtm) ? MSB(sqs) : LSB(sqs))
#  define LeastAdvanced(wtm, sqs) ((wtm) ? LSB(sqs) : MSB(sqs))
#  define MinMax(wtm, v1, v2)     ((wtm) ? Min((v1), (v2)) : Max((v1), (v2)))
#  define FrontOf(wtm, k, p)      ((wtm) ? k > p : k < p)
#  define Behind(wtm, k, p)       ((wtm) ? k < p : k > p)
#  define AttacksRank(a)          (AttacksRook(a, OccupiedSquares) & rank_mask[Rank(a)])
#  define AttacksFile(a)          (AttacksRook(a, OccupiedSquares) & file_mask[File(a)])
#  define AttacksDiaga1(a)        (AttacksBishop(a, OccupiedSquares) & (plus9dir[a] | minus9dir[a]))
#  define AttacksDiagh1(a)        (AttacksBishop(a, OccupiedSquares) & (plus7dir[a] | minus7dir[a]))
#  define Passed(sq, wtm)         (!(mask_pawn_passed[wtm][sq] & Pawns(Flip(wtm))))
/*
   the following macros are used to extract the pieces of a move that are
   kept compressed into the rightmost 21 bits of a simple integer.
 */
#  define From(a)                 ((a)&63)
#  define To(a)                   (((a)>>6)&63)
#  define Piece(a)                (((a)>>12)&7)
#  define Captured(a)             (((a)>>15)&7)
#  define Promote(a)              (((a)>>18)&7)
#  define CaptureOrPromote(a)     (((a)>>15)&63)
#  define Castling(a)             ((Piece(a) == king) && (Abs(From(a) - To(a)) == 2))
#  define SetMask(a)              (set_mask[a])
#  define ClearMask(a)            (clear_mask[a])
#  define Pawns(c)                (tree->pos.color[c].pieces[pawn])
#  define Knights(c)              (tree->pos.color[c].pieces[knight])
#  define Bishops(c)              (tree->pos.color[c].pieces[bishop])
#  define Rooks(c)                (tree->pos.color[c].pieces[rook])
#  define Queens(c)               (tree->pos.color[c].pieces[queen])
#  define Kings(c)                (tree->pos.color[c].pieces[king])
#  define KingSQ(c)               (tree->pos.kingsq[c])
#  define Occupied(c)             (tree->pos.color[c].pieces[occupied])
#  define Pieces(c, p)            (tree->pos.color[c].pieces[p])
#  define TotalPieces(c, p)       (tree->pos.pieces[c][p])
#  define PieceValues(c, p)       (piece_values[c][p])
#  define TotalAllPieces          (tree->pos.total_all_pieces)
#  define Material                (tree->pos.material_evaluation)
#  define MaterialRoot            (root_material_evaluation)
#  define MaterialSTM(side)       ((side) ? Material : -Material)
#  define MaterialRootSTM(side)   ((side) ? MaterialRoot : -MaterialRoot)
#  define Castle(ply, c)          (tree->position[ply].castle[c])
#  define Rule50Moves(ply)        (tree->position[ply].rule_50_moves)
#  define Repetition(side)        (tree->rep_index[side])
#  define HashKey                 (tree->pos.hash_key)
#  define PawnHashKey             (tree->pos.pawn_hash_key)
#  define EnPassant(ply)          (tree->position[ply].enpassant_target)
#  define EnPassantTarget(ply)    (EnPassant(ply) ? SetMask(EnPassant(ply)) : 0)
#  define PcOnSq(sq)              (tree->pos.board[sq])
#  define BishopsQueens           (tree->pos.bishops_queens)
#  define RooksQueens             (tree->pos.rooks_queens)
#  define OccupiedSquares         (Occupied(white) | Occupied(black))
#  define Color(square)           (square_color[square] ? dark_squares : ~dark_squares)
#  define SideToMove(c)           ((c) ? "White" : "Black")
#  define CanWin(c)               ((c) ? (can_win & CAN_WIN_WHITE) : (can_win & CAN_WIN_BLACK))
#  define SetCannotWin(c)         ((c) ? (can_win &= CANNOT_WIN_WHITE) : (can_win &= CANNOT_WIN_BLACK))
#  define Reduction(n, l, d, m)(reductions[n == PV_NODE ? 1 : 0][l ? 1 : 0][Min(d, 31)][Min(m, 63)])
#  define Moves(d)                (movecnt_margin[Min(d, 7)])
#  define CPUPercent              Min(((program_end_time - program_start_time > 0) ? ((double)(program_end_cpu_time - \
                                    program_start_cpu_time) * 100.0 / (double)(program_end_time - program_start_time) \
                                    / (double)Max(1, smp_max_threads)) : 0.0), 100.0)
#  define Resourceful(c)          (TotalPieces(c, occupied) > p_vals[queen])
#  define NotMate(alpha, beta)    (alpha >= -NOT_MATE && beta <= NOT_MATE)
#  define MaterialBalanceRoot(pv) (Abs(Material - MaterialRoot) < pv)
#  define DoLazy(lscore, cutoff)  (lscore + cutoff < alpha || lscore - cutoff > beta)
#  define NodeType(ply)           ((node_type == CUT_NODE) ? ((tree->phase[ply] >= HISTORY_MOVES) ? ALL_NODE : node_type) : node_type)
#  define DoFutility(ply)         (tree->phase[ply] >= HISTORY_MOVES)
#  define DoDelta(ply)            (tree->phase[ply] >= HISTORY_MOVES)
#  define DoLateFutility(ply)     (tree->phase[ply] >= HISTORY_MOVES)
#  define DoReduction(ply)        (tree->phase[ply] >= HISTORY_MOVES)
/*
   the following macros are used to Set and Clear a specific bit in the
   second argument.  this is done to make the code more readable, rather
   than to make it faster.
 */
#  define ClearSet(a,b)           b=((a)^(b))
#  define Clear(a,b)              b=ClearMask(a)&(b)
#  define Set(a,b)                b=SetMask(a)|(b)
/*
   the following macros are used to update the hash signatures.
 */
#  define Hash(stm,piece,square)     (HashKey^=randoms[stm][piece][square])
#  define HashP(stm,square)          (PawnHashKey^=randoms[stm][pawn][square])
#  define HashCastle(stm,direction)  (HashKey^=castle_random[stm][direction])
#  define HashEP(stm)                (HashKey^=enpassant_random[stm])
#  define SavePV(tree,ply,ph)   do {                                      \
        tree->pv[ply - 1].path[ply - 1] = tree->curmv[ply - 1];           \
        tree->pv[ply - 1].pathl = ply;                                    \
        tree->pv[ply - 1].pathh = ph;                                     \
        tree->pv[ply - 1].pathd = iteration_depth;                        \
        } while(0)
#  if defined(INLINE64)
#    include "inline64.h"
#  endif
#  if defined(INLINE32)
#    include "inline32.h"
#  endif
#  if defined(UNIX)
#    define SPEAK "./speak "
#  else
#    define SPEAK ".\\Speak.exe "
#  endif
#endif                          /* if defined(TYPES_INCLUDED) */
