#include "chess.h"
#include "data.h"
#define BIG_SORT_VALUE 999999
#define BIG_HISTORY_VALUE 99999999
#define HISTORY_THRESHOLD 1000

/* last modified 05/25/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   NextMove() is used to select the next move from the current move list.    *
 *                                                                             *
 *******************************************************************************
 */
int NextMove(TREE * RESTRICT tree, int alpha, int beta, int ply, int wtm,
  int depth, NODE_TYPE node_type, int *best_capture)
{
  register int *movep, *sortv, t, *bestp, bests, bestval, hvalue;

  switch (tree->next_status[ply].phase) {
/*
 ************************************************************
 *                                                          *
 *   first, try the transposition table move (which will be *
 *   the principal variation move as we first move down the *
 *   tree).                                                 *
 *                                                          *
 ************************************************************
 */
  case HASH_MOVE:
    tree->next_status[ply].excluded = 0;
    tree->next_status[ply].phase = GENERATE_CAPTURE_MOVES;
    if (tree->hash_move[ply]) {
      sortv = tree->sort_value;
      *sortv = 0;
      t = 0;
      tree->curmv[ply] = tree->hash_move[ply];
      if (Captured(tree->curmv[ply])) {
        *sortv = MVV_LVA[Captured(tree->curmv[ply])][Piece(tree->curmv[ply])];
        if (*sortv < 0 && (t = Swap(tree, tree->curmv[ply], wtm)) >= 0)
          *sortv = t;
      }
      if (Promote(tree->curmv[ply]) && t == 0) {
        register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
        register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
        if (!loose)
          *sortv += pc_values[Promote(tree->curmv[ply])] -
            (Captured(tree->curmv[ply]) ? 0 : pc_values[pawn]);
      }
      if (*best_capture < *sortv + Piece(tree->curmv[ply]) && t == 0) {
        register BITBOARD attacks = AttacksTo(tree, To(tree->curmv[ply]));
        register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
          if (loose)
            *best_capture = Max(*best_capture, *sortv);
          else
            *best_capture = Max(*best_capture, *sortv +
              ((Captured(tree->curmv[ply])) ? Piece(tree->curmv[ply]) : 0));
      }
      tree->next_status[ply].done[tree->next_status[ply].excluded++] =
        tree->curmv[ply];
      return (HASH_MOVE);
    }
/*
 ************************************************************
 *                                                          *
 *   generate captures and sort them based on (a) the value *
 *   of the captured piece - the value of the capturing     *
 *   piece if this is > 0; or, (b) the value returned by    *
 *   Swap().                                                *
 *                                                          *
 ************************************************************
 */
  case GENERATE_CAPTURE_MOVES:
    tree->next_status[ply].phase = CAPTURE_MOVES;
    tree->last[ply] = GenerateCaptures(tree, ply, wtm, tree->last[ply - 1]);
/*
 ************************************************************
 *                                                          *
 *  Now make a pass over the moves to assign the sort value *
 *  for each.  We simply use MVV/LVA move order here.  A    *
 *  simple optimization is to use the pre-computed array    *
 *  MVV_LVA[victim][attacker] which returns a value that    *
 *  indicates MVV/LVA order.                                *
 *                                                          *
 ************************************************************
 */
    tree->next_status[ply].captures = 0;
    if (tree->last[ply] > tree->last[ply - 1]) {
      for (movep = tree->last[ply - 1], sortv = tree->sort_value;
          movep < tree->last[ply]; movep++, sortv++) {
        if (tree->hash_move[ply] && *movep == tree->hash_move[ply]) {
          *sortv = BIG_SORT_VALUE;
          tree->next_status[ply].captures++;
        } else {
          *sortv = 0;
          t = 0;
          if (Captured(*movep)) {
            *sortv = MVV_LVA[Captured(*movep)][Piece(*movep)];
            if (*sortv < 0 && (t = Swap(tree, *movep, wtm)) >= 0)
              *sortv = t;
          }
          if (Promote(*movep) && t == 0) {
            register BITBOARD attacks = AttacksTo(tree, To(*movep));
            register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
            if (!loose)
              *sortv += pc_values[Promote(*movep)] -
                (Captured(*movep) ? 0 : pc_values[pawn]);
          }
          if (*best_capture < *sortv + Piece(*movep) && t == 0) {
            register BITBOARD attacks = AttacksTo(tree, To(*movep));
            register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
            if (loose)
              *best_capture = Max(*best_capture, *sortv);
            else
              *best_capture = Max(*best_capture, *sortv +
                ((Captured(*movep)) ? Piece(*movep) : 0));
          }
          if (*sortv >= 0)
            tree->next_status[ply].captures++;
        }
      }
    }
/*
 ************************************************************
 *                                                          *
 *  This is a simple insertion sort algorithm.  It seems to *
 *  be no faster than a normal bubble sort, but using this  *
 *  eliminated a lot of explaining about "why?". :)         *
 *                                                          *
 ************************************************************
 */
    if (tree->last[ply] > tree->last[ply - 1] + 1) {
      register int move1, sortv1, *pmove0, *psortv0;
      sortv = tree->sort_value + 1;
      for (movep = tree->last[ply - 1] + 1; movep < tree->last[ply]; movep++, sortv++) {
        move1 = *movep;
        sortv1 = *sortv;
        pmove0 = movep - 1;
        psortv0 = sortv - 1;
        while (pmove0 >= tree->last[ply - 1] && ((*psortv0 < sortv1) || (*psortv0 == sortv1 &&
          *pmove0 && move1 && pc_values[Captured(*pmove0)] < pc_values[Captured(move1)]))) {
          *(psortv0 + 1) = *psortv0;
          *(pmove0 + 1) = *pmove0;
          pmove0--;
          psortv0--;
        }
        *(pmove0 + 1) = move1;
        *(psortv0 + 1) = sortv1;
      }
    }
    tree->next_status[ply].last = tree->last[ply - 1];
/*
 ************************************************************
 *                                                          *
 *   try the captures moves, which are in order based on    *
 *   the expected gain of material.  captures that lose     *
 *   material have been excluded from this phase.           *
 *                                                          *
 ************************************************************
 */
  case CAPTURE_MOVES:
    while (tree->next_status[ply].captures) {
      tree->curmv[ply] = *(tree->next_status[ply].last++);
      tree->next_status[ply].captures--;
      if (!tree->next_status[ply].captures)
        tree->next_status[ply].phase = KILLER_MOVE_1;
      if (Exclude(tree, ply, tree->curmv[ply])) { /* if hash move */
        continue;
      }
      *(tree->next_status[ply].last - 1) = NULL_MOVE;
      return (CAPTURE_MOVES);
    }
/*
 ************************************************************
 *                                                          *
 *  Now, try the killer moves.  This phase tries the two    *
 *  killers for the current ply without generating moves,   *
 *  which saves time if a cutoff occurs.  After those two   *
 *  killers are searched, we try the killers from two plies *
 *  back since they have greater depth and might produce a  *
 *  cutoff if the current two do not.                       *
 *                                                          *
 ************************************************************
 */
  case KILLER_MOVE_1:
    Lock(lock_his);
    if (!Exclude(tree, ply, tree->killers[ply].move1) &&
        ValidMove(tree, ply, wtm, tree->killers[ply].move1)) {
      tree->curmv[ply] = tree->killers[ply].move1;
      tree->next_status[ply].done[tree->next_status[ply].excluded++] =
        tree->curmv[ply];
      tree->next_status[ply].phase = KILLER_MOVE_2;
      Unlock(lock_his);
      return (KILLER_MOVE_1);
    }
    Unlock(lock_his);
  case KILLER_MOVE_2:
    Lock(lock_his);
    if (!Exclude(tree, ply, tree->killers[ply].move2) &&
        ValidMove(tree, ply, wtm, tree->killers[ply].move2)) {
      tree->curmv[ply] = tree->killers[ply].move2;
      tree->next_status[ply].done[tree->next_status[ply].excluded++] =
        tree->curmv[ply];
      tree->next_status[ply].phase =
          (ply > 2) ? KILLER_MOVE_3 : ((ply > 1) ? COUNTER_MOVE1 : BAD_CAPTURE_MOVES);
      Unlock(lock_his);
      return (KILLER_MOVE_2);
    }
    Unlock(lock_his);
  case KILLER_MOVE_3:
    Lock(lock_his);
    if (!Exclude(tree, ply, tree->killers[ply - 2].move1) &&
        ValidMove(tree, ply, wtm, tree->killers[ply - 2].move1)) {
      tree->curmv[ply] = tree->killers[ply - 2].move1;
      tree->next_status[ply].done[tree->next_status[ply].excluded++] =
        tree->curmv[ply];
      tree->next_status[ply].phase = KILLER_MOVE_4;
      Unlock(lock_his);
      return (KILLER_MOVE_3);
    }
    Unlock(lock_his);
  case KILLER_MOVE_4:
    Lock(lock_his);
    if (!Exclude(tree, ply, tree->killers[ply - 2].move2) &&
        ValidMove(tree, ply, wtm, tree->killers[ply - 2].move2)) {
      tree->curmv[ply] = tree->killers[ply - 2].move2;
      tree->next_status[ply].done[tree->next_status[ply].excluded++] =
        tree->curmv[ply];
      tree->next_status[ply].phase =
        (ply > 1) ? COUNTER_MOVE1 : BAD_CAPTURE_MOVES;
      Unlock(lock_his);
      return (KILLER_MOVE_4);
    }
    Unlock(lock_his);
/*
 ************************************************************
 *                                                          *
 *  Now, before we give up and generate moves, try the      *
 *  counter-move which was a move that failed high in the   *
 *  past when the move at the previous ply was played.      *
 *                                                          *
 ************************************************************
 */
  case COUNTER_MOVE1:
    Lock(lock_his);
    if (!Exclude(tree, ply, counter_move[tree->curmv[ply - 1] & 4095].move1) &&
        ValidMove(tree, ply, wtm, counter_move[tree->curmv[ply - 1] & 4095].move1)) {
      tree->curmv[ply] = counter_move[tree->curmv[ply - 1] & 4095].move1;
      tree->next_status[ply].done[tree->next_status[ply].excluded++] =
        tree->curmv[ply];
      tree->next_status[ply].phase = COUNTER_MOVE2;
      Unlock(lock_his);
      return (COUNTER_MOVE1);
    }
    Unlock(lock_his);
  case COUNTER_MOVE2:
    Lock(lock_his);
    if (!Exclude(tree, ply, counter_move[tree->curmv[ply - 1] & 4095].move2) &&
        ValidMove(tree, ply, wtm, counter_move[tree->curmv[ply - 1] & 4095].move2)) {
      tree->curmv[ply] = counter_move[tree->curmv[ply - 1] & 4095].move2;
      tree->next_status[ply].done[tree->next_status[ply].excluded++] =
        tree->curmv[ply];
      tree->next_status[ply].phase =
        (ply > 2) ? MOVE_PAIR1 : BAD_CAPTURE_MOVES;
      Unlock(lock_his);
      return (COUNTER_MOVE2);
    }
    Unlock(lock_his);
  case MOVE_PAIR1:
    Lock(lock_his);
    if (!Exclude(tree, ply, move_pair[tree->curmv[ply - 2] & 4095].move1) &&
        ValidMove(tree, ply, wtm, move_pair[tree->curmv[ply - 2] & 4095].move1)) {
      tree->curmv[ply] = move_pair[tree->curmv[ply - 2] & 4095].move1;
      tree->next_status[ply].done[tree->next_status[ply].excluded++] =
        tree->curmv[ply];
      tree->next_status[ply].phase = MOVE_PAIR2;
      Unlock(lock_his);
      return (MOVE_PAIR1);
    }
    Unlock(lock_his);
  case MOVE_PAIR2:
    Lock(lock_his);
    if (!Exclude(tree, ply, move_pair[tree->curmv[ply - 2] & 4095].move2) &&
        ValidMove(tree, ply, wtm, move_pair[tree->curmv[ply - 2] & 4095].move2)) {
      tree->curmv[ply] = move_pair[tree->curmv[ply - 2] & 4095].move2;
      tree->next_status[ply].done[tree->next_status[ply].excluded++] =
        tree->curmv[ply];
      tree->next_status[ply].phase = BAD_CAPTURE_MOVES;
      Unlock(lock_his);
      return (MOVE_PAIR2);
    }
    Unlock(lock_his);
/*
************************************************************
*                                                          *
*   many programmers favour losing captures before other   *
*   none-captures, directly behind the killers.  they are  *
*   kind of forced, and one possibly has to deal with all  *
*   kind of tactical motives and interactions, one may not *
*   consider in move ordering, such as pins, batteries,    *
*   discovered attacks and overloaded defenders.           *
*   otherwise, obviously losing captures are likely        *
*   refuted cheaply.  but if a losing capture fails high   *
*   for some reason, we have saved the effort to generate, *
*   and more importantly to search other non-captures at   *
*   all.                                                   *
*                                                          *
************************************************************
*/
  case BAD_CAPTURE_MOVES:
    tree->next_status[ply].phase = BAD_CAPTURE_MOVES;
    tree->next_status[ply].last = tree->last[ply - 1];
    for (; tree->next_status[ply].last < tree->last[ply];
        tree->next_status[ply].last++) {
      if (*(tree->next_status[ply].last)) { /* if not NULL_MOVE */
        if (!Exclude(tree, ply, *(tree->next_status[ply].last))) { /* if not hash, killer, counter or pair move */
          tree->curmv[ply] = *(tree->next_status[ply].last);
          tree->next_status[ply].last++;
          *(tree->next_status[ply].last - 1) = NULL_MOVE;
          return (BAD_CAPTURE_MOVES);
        }
      }
    }
/*
 ************************************************************
 *                                                          *
 *   now, generate all non-capturing moves and sort them    *
 *   based on the value returned by Move().                 *
 *                                                          *
 ************************************************************
 */
  case GENERATE_QUIET_MOVES:
    tree->next_status[ply].phase = REMAINING_MOVES;
    tree->last[ply] = GenerateNoncaptures(tree, ply, wtm, tree->last[ply]);
    tree->next_status[ply].remaining = 0;
    if (tree->last[ply] > tree->last[ply - 1]) {
      for (movep = tree->last[ply - 1], sortv = tree->sort_value;
          movep < tree->last[ply]; movep++, sortv++) {
        if (*movep) { /* if not NULL_MOVE */
          if (Exclude(tree, ply, *movep)) /* if hash, killer, counter or pair move */
            *sortv = 0;
          else if (!Captured(*movep)) {
            *sortv = Move(tree, alpha, beta, ply, *movep, wtm, node_type, *best_capture);
            if (Promote(*movep)) {
              register BITBOARD attacks = AttacksTo(tree, To(*movep));
              register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
              if (!loose)
                *sortv += pc_values[Promote(*movep)];
            }
          }
          else
            *sortv = 0;
          if (!Exclude(tree, ply, *movep)) { /* if not hash, killer, counter or pair move */
            if (*sortv + late_moves_margin > 0) /* not in LATE_MOVES */
              tree->next_status[ply].remaining++;
          }
        }
        else
          *sortv = 0;
      }
    }
/*
 ************************************************************
 *                                                          *
 *  now, try the history moves.  this phase takes the       *
 *  complete move list, and passes over them in a classic   *
 *  selection-sort, choosing the move with the highest      *
 *  history score.  this phase is only done one time, as it *
 *  also purges the hash, killer, counter and paired moves  *
 *  from the list.                                          *
 *                                                          *
 ************************************************************
 */
     if (tree->last[ply] > tree->last[ply - 1]) {
      if (depth >= 6) {
        tree->next_status[ply].history = 0;
        tree->next_status[ply].phase = HISTORY_MOVES;
        bestval = -BIG_HISTORY_VALUE;
        bestp = NULL;
        bests = 0;
        Lock(lock_his);
        sortv = tree->sort_value;
        for (movep = tree->last[ply - 1]; movep < tree->last[ply]; movep++, sortv++)
          if (*movep) { /* if not NULL_MOVE */
            if (Exclude(tree, ply, *movep))
              *movep = NULL_MOVE; /* exclude hash, killer, counter and pair moves */
            else {
              hvalue = history[wtm][HistoryIndex(*movep)];
              if (hvalue > bestval) {
                bestval = hvalue;
                bestp = movep;
                bests = *sortv;
              }
            }
          }
        Unlock(lock_his);
        if (bestp) {
          tree->curmv[ply] = *bestp;
          if (tree->next_status[ply].remaining) {
            if (bests + late_moves_margin > 0) /* not in LATE_MOVES */
              tree->next_status[ply].remaining--;
          }
          *bestp = NULL_MOVE;
          if (bestval < -HISTORY_THRESHOLD)
            tree->phase[ply] = REMAINING_MOVES;
          return (HISTORY_MOVES);
        }
        goto remaining_moves;
      }
      else
        goto remaining_moves;
    }
    else
      goto remaining_moves;
/*
 ************************************************************
 *                                                          *
 *  now, continue with the history moves, but since one     *
 *  pass has been made over the complete move list, there   *
 *  are no hash, killer, counter or paired moves left in    *
 *  the list, so the tests for these can be avoided.        *
 *                                                          *
 ************************************************************
 */
  case HISTORY_MOVES:
    if (tree->last[ply] > tree->last[ply - 1]) {
      if (depth >= 6) {
        bestval = -BIG_HISTORY_VALUE;
        bestp = NULL;
        bests = 0;
        Lock(lock_his);
        sortv = tree->sort_value;
        for (movep = tree->last[ply - 1]; movep < tree->last[ply]; movep++, sortv++)
          if (*movep) { /* if not NULL_MOVE */
            hvalue = history[wtm][HistoryIndex(*movep)];
            if (hvalue > bestval) {
              bestval = hvalue;
              bestp = movep;
              bests = *sortv;
            }
          }
        Unlock(lock_his);
        if (bestp) {
          tree->curmv[ply] = *bestp;
          if (tree->next_status[ply].remaining) {
            if (bests + late_moves_margin > 0) /* not in LATE_MOVES */
              tree->next_status[ply].remaining--;
          }
          *bestp = NULL_MOVE;
          if (bestval < -HISTORY_THRESHOLD)
            tree->phase[ply] = REMAINING_MOVES;
          tree->next_status[ply].history++;
          if (tree->next_status[ply].history >= history_max_moves) {
            tree->next_status[ply].phase = REMAINING_MOVES;
            tree->next_status[ply].last = tree->last[ply - 1];
          }
          return (HISTORY_MOVES);
        }
      }
      else
        goto remaining_moves;
    }
  remaining_moves:
/*
 ************************************************************
 *                                                          *
 *  This is a simple insertion sort algorithm.  It seems to *
 *  be no faster than a normal bubble sort, but using this  *
 *  eliminated a lot of explaining about "why?". :)         *
 *                                                          *
 ************************************************************
 */
    if (tree->last[ply] > tree->last[ply - 1] + 1) {
      register int move1, sortv1, *pmove0, *psortv0;
      sortv = tree->sort_value + 1;
      for (movep = tree->last[ply - 1] + 1; movep < tree->last[ply]; movep++, sortv++) {
        *sortv = (*movep == NULL_MOVE) ? 0 : *sortv;
        *(sortv - 1) = (*(movep - 1) == NULL_MOVE) ? 0 : *(sortv - 1);
        move1 = *movep;
        sortv1 = *sortv;
        pmove0 = movep - 1;
        psortv0 = sortv - 1;
        while (pmove0 >= tree->last[ply - 1] && ((*psortv0 < sortv1) || (*psortv0 == sortv1 &&
          *pmove0 && move1 && pc_values[Captured(*pmove0)] < pc_values[Captured(move1)]))) {
          *(psortv0 + 1) = *psortv0;
          *(pmove0 + 1) = *pmove0;
          pmove0--;
          psortv0--;
        }
        *(pmove0 + 1) = move1;
        *(psortv0 + 1) = sortv1;
      }
    }
    tree->next_status[ply].last = tree->last[ply - 1];
    tree->next_status[ply].phase = REMAINING_MOVES;
/*
 ************************************************************
 *                                                          *
 *  then we try the rest of the set of moves, and we do not *
 *  use Exclude() function to skip any moves we have        *
 *  already searched (hash or killers) since the history    *
 *  phase above has already done that.                      *
 *                                                          *
 ************************************************************
 */
  case REMAINING_MOVES:
  case LATE_MOVES:
    for (; tree->next_status[ply].last < tree->last[ply];
        tree->next_status[ply].last++) {
      if (*(tree->next_status[ply].last)) { /* if not NULL_MOVE */
        if (tree->next_status[ply].phase == REMAINING_MOVES) {
          if (tree->next_status[ply].remaining) {
            tree->next_status[ply].remaining--;
            if (!tree->next_status[ply].remaining)
              tree->next_status[ply].phase = LATE_MOVES;
          } else
            tree->next_status[ply].phase = LATE_MOVES;
        }
        tree->curmv[ply] = *(tree->next_status[ply].last);
        tree->next_status[ply].last++;
        return (tree->next_status[ply].phase);
      }
    }
    return (NONE);
  default:
    Print(4095, "oops!  next_status.phase is bad! [normal %d]\n",
        tree->next_status[ply].phase);
    return (NONE);
  }
}

/* last modified 10/21/14 */
/*
 *******************************************************************************
 *                                                                             *
 *   Exclude() searches the list of moves searched prior to generating a move  *
 *   list to exclude those that were searched via a hash table best move or    *
 *   through the killer moves for the current ply and two plies back.          *
 *                                                                             *
 *   The variable next_status[].excluded is the total number of non-generated  *
 *   moves we searched.  next_status[].remain is initially set to excluded,    *
 *   but each time an excluded move is found, the counter is decremented.      *
 *   Once all excluded moves have been found, we avoid running through the     *
 *   list of excluded moves on each call and simply return.                    *
 *                                                                             *
 *******************************************************************************
 */
int Exclude(TREE * RESTRICT tree, int ply, int move) {
  int i;

  if (move == NULL_MOVE)
    return 1;
  if (tree->next_status[ply].excluded)
    for (i = 0; i < tree->next_status[ply].excluded; i++)
      if (move == tree->next_status[ply].done[i]) {
        return 1;
      }
  return 0;
}
