#include "chess.h"
#include "data.h"
#define BIG_SORT_VALUE 999999
/* last modified 05/15/22 */
/*
 *******************************************************************************
 *                                                                             *
 *   NextEvasion() is used to select the next move from the current move list  *
 *   when the king is in check.  we use GenerateEvasions() (in movgen.c) to    *
 *   generate a list of moves that get us out of check.  The only unusual      *
 *   feature is that these moves are all legal and do not need to be vetted    *
 *   with the usual Check() function to test for legality.                     *
 *                                                                             *
 *******************************************************************************
 */
int NextEvasion(TREE * RESTRICT tree, int ply, int wtm)
{
  register int *movep, *sortv, t;

  switch (tree->next_status[ply].phase) {
/*
 ************************************************************
 *                                                          *
 *   first generate all legal moves by using the special    *
 *   GenerateCheckEvasions() function, so we can determine  *
 *   if this is a one-legal-reply-to-check position.        *
 *                                                          *
 ************************************************************
 */
/*
 ************************************************************
 *                                                          *
 *   first, try the transposition table move (which will be *
 *   the principal variation move as we first move down the *
 *   tree).                                                 *
 *                                                          *
 ************************************************************
 */
  case HASH_MOVE:
  tree->next_status[ply].phase = GENERATE_QUIET_MOVES;
    if (tree->hash_move[ply]) {
      tree->curmv[ply] = tree->hash_move[ply];
      return (HASH_MOVE);
    }
/*
 ************************************************************
 *                                                          *
 *   generate evasions and sort them based on (a) the value *
 *   of the captured piece - the value of the capturing     *
 *   piece if this is > 0; or, (b) the value returned by    *
 *   Swap().                                                *
 *                                                          *
 ************************************************************
 */
  case GENERATE_QUIET_MOVES:
    tree->next_status[ply].phase = REMAINING_MOVES;
    tree->last[ply] =
        GenerateCheckEvasions(tree, ply, wtm, tree->last[ply - 1]);
    if (tree->last[ply] > tree->last[ply - 1]) {
      for (movep = tree->last[ply - 1], sortv = tree->sort_value;
          movep < tree->last[ply]; movep++, sortv++) {
        if (tree->hash_move[ply] && *movep == tree->hash_move[ply]) {
          *sortv = BIG_SORT_VALUE;
        } else {
          *sortv = 0;
          t = 0;
          if (Captured(*movep)) {
            *sortv = MVV_LVA[Captured(*movep)][Piece(*movep)];
            if (*sortv < 0 && (t = Swap(tree, *movep, wtm)) >= 0)
              *sortv = t;
          }
          if (Promote(*movep) && t == 0) {
            register BITBOARD attacks = AttacksTo(tree, To(*movep));
            register int loose = (attacks & (attacks & Occupied(Flip(wtm)))) ? 1 : 0;
            if (!loose)
              *sortv += pc_values[Promote(*movep)] -
                (Captured(*movep) ? 0 : pc_values[pawn]);
          }
        }
      }
    }
/*
 ************************************************************
 *                                                          *
 *  This is a simple insertion sort algorithm.  It seems to *
 *  be no faster than a normal bubble sort, but using this  *
 *  eliminated a lot of explaining about "why?". :)         *
 *                                                          *
 ************************************************************
 */
    if (tree->last[ply] > tree->last[ply - 1] + 1) {
      register int move1, sortv1, *pmove0, *psortv0;
      sortv = tree->sort_value + 1;
      for (movep = tree->last[ply - 1] + 1; movep < tree->last[ply]; movep++, sortv++) {
        move1 = *movep;
        sortv1 = *sortv;
        pmove0 = movep - 1;
        psortv0 = sortv - 1;
        while (pmove0 >= tree->last[ply - 1] && ((*psortv0 < sortv1) || (*psortv0 == sortv1 &&
          *pmove0 && move1 && pc_values[Captured(*pmove0)] < pc_values[Captured(move1)]))) {
          *(psortv0 + 1) = *psortv0;
          *(pmove0 + 1) = *pmove0;
          pmove0--;
          psortv0--;
        }
        *(pmove0 + 1) = move1;
        *(psortv0 + 1) = sortv1;
      }
    }
    tree->next_status[ply].last = tree->last[ply - 1];
/*
 ************************************************************
 *                                                          *
 *   now try the rest of the set of moves.                  *
 *                                                          *
 ************************************************************
 */
  case REMAINING_MOVES:
    for (; tree->next_status[ply].last < tree->last[ply];
        tree->next_status[ply].last++) {
      if ((*tree->next_status[ply].last) &&
        *(tree->next_status[ply].last) != tree->hash_move[ply]) {
        tree->curmv[ply] = *(tree->next_status[ply].last);
        *tree->next_status[ply].last++ = NULL_MOVE;
        return (REMAINING_MOVES);
      }
    }
    return (NONE);
  default:
    printf("oops!  next_status.phase is bad! [evasion %d]\n",
        tree->next_status[ply].phase);
    return (NONE);
  }
}
