#include "chess.h"
#include "data.h"
/* last modified 12/22/21 */
/*
 *******************************************************************************
 *                                                                             *
 *   NextRootMove() is used to select the next move from the root move list.   *
 *                                                                             *
 *******************************************************************************
 */
int NextRootMove(TREE * RESTRICT tree, TREE * RESTRICT mytree, int wtm,
  int *best_capture)
{
  register int done, which, i;
  UINT64 total_nodes;

  *best_capture = best_root_capture;
/*
 ************************************************************
 *                                                          *
 *   for the moves at the root of the tree, the list has    *
 *   already been generated and sorted.  on entry, test     *
 *   the searched_this_root_move[] array and then take the  *
 *   moves in the order they appear in the move list.       *
 *                                                          *
 ************************************************************
 */
  time_abort += TimeCheck(tree, 0);
  if (time_abort)
    return (NONE);
  if (!annotate_mode && !pondering && !booking && n_root_moves == 1 &&
      iteration_depth > 4) {
    abort_search = 1;
    return (NONE);
  }
  done = 0;
  for (which = 0; which < n_root_moves; which++)
    if (root_moves[which].status & 16)
      done++;
  Lock(lock_root);
  if (done == 1 && (root_moves[0].status & 16) && root_value == root_alpha &&
      !(root_moves[0].status & 2)) {
    Unlock(lock_root);
    return (NONE);
  }
  Unlock(lock_root);
  for (which = 0; which < n_root_moves; which++)
    if (!(root_moves[which].status & 16)) {
      if (search_move) {
        if (root_moves[which].move != search_move) {
          root_moves[which].status |= 16;
          continue;
        }
      }
      tree->curmv[1] = root_moves[which].move;
      tree->root_move = which;
      root_moves[which].status |= 16;
      if ((ReadClock() - start_time > noise_level) && (display_options & 32)) {
        Lock(lock_io);
        sprintf(mytree->remaining_moves_text, "%d/%d", which + 1, n_root_moves);
        end_time = ReadClock();
        if (pondering)
          printf("               %2i   %s%7s?  ", iteration_depth,
              DisplayTime(end_time - start_time), mytree->remaining_moves_text);
        else
          printf("               %2i   %s%7s*  ", iteration_depth,
              DisplayTime(end_time - start_time), mytree->remaining_moves_text);
        if (display_options & 32 && display_options & 64)
          printf("%d. ", move_number);
        if ((display_options & 32) && (display_options & 64) && Flip(wtm))
          printf("... ");
        strcpy(mytree->root_move_text, OutputMove(tree, tree->curmv[1], 1,
                wtm));
        total_nodes = block[0]->nodes_searched;
        for (i = 1; i < MAX_BLOCKS; i++)
          if (block[i] && block[i]->used)
            total_nodes += block[i]->nodes_searched;
        nodes_per_second = total_nodes * 100 / Max(end_time - start_time, 1);
        i = strlen(mytree->root_move_text);
        i = (i < 8) ? i : 8;
        strncat(mytree->root_move_text, "          ", 8 - i);
        printf("%s", mytree->root_move_text);
        printf("(%snps)             \r", DisplayKM(nodes_per_second));
        fflush(stdout);
        Unlock(lock_io);
      }
/*
 ************************************************************
 *                                                          *
 *   Bit of a tricky exit.  If the move is flagged as "do   *
 *   not reduce" or "do not search in parallel" then we     *
 *   return "HASH_MOVE" which will prevent Search() from    *
 *   reducing the move (LMR).  Otherwise we return the more *
 *   common "HISTORY_MOVES" value which allows LMR to       *
 *   be used on those root moves.                           *
 *                                                          *
 ************************************************************
 */
      if (root_moves[which].status & (4 + 8))
        return (HASH_MOVE);
      else
        return (HISTORY_MOVES);
    }
  return (NONE);
}

/* last modified 08/07/05 */
/*
 *******************************************************************************
 *                                                                             *
 *   NextRootMoveParallel() is used to determine if the next root move can be  *
 *   searched in parallel.  if it appears to Iterate() that one of the moves   *
 *   following the first move might become the best move, the 'no parallel'    *
 *   flag is set to speed up finding the new best move.                        *
 *                                                                             *
 *******************************************************************************
 */
int NextRootMoveParallel(void)
{
  register int aspiration = aspiration_window;
  register int which;

/*
 ************************************************************
 *                                                          *
 *   first, find out how far down the list we have searched *
 *   already.  if the next move is flagged as "do not       *
 *   search in parallel" then return 1 unless the score has *
 *   dropped significantly.  if the score has dropped, then *
 *   we search serially to find a better move quickly.      *
 *                                                          *
 ************************************************************
 */
  for (which = 0; which < n_root_moves; which++)
    if (!(root_moves[which].status & 16))
      break;
  if (which < n_root_moves && root_moves[which].status & 4)
    return (0);
  Lock(lock_root);
  if (root_value >= last_root_value - aspiration || which > n_root_moves / 3) {
    Unlock(lock_root);
    return (1);
  }
  Unlock(lock_root);
  return (0);
}
