#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "chess.h"
#include "data.h"

/* last modified 06/19/19 */
/*
 *******************************************************************************
 *                                                                             *
 *   Test() is used to test the program against a suite of test positions to   *
 *   measure its performance on a particular machine, or to evaluate its skill *
 *   after modifying it in some way.                                           *
 *                                                                             *
 *   the test is initiated by using the "test <filename>" command to read in   *
 *   the suite of problems from file <filename>.  the format of this file is   *
 *   as follows:                                                               *
 *                                                                             *
 *   setboard <forsythe-string>:  this sets the board position using the usual *
 *   forsythe notation (see module SetBoard() in setc for a full ex-           *
 *   planation of the syntax).                                                 *
 *                                                                             *
 *   solution <move1> <move2> ... <moven>:  this provides a solution move (or  *
 *   set of solution moves if more than one is correct).  if the search finds  *
 *   one of these moves, then the prblem is counted as correct, otherwise it   *
 *   is counted wrong.                                                         *
 *                                                                             *
 *   after reading these two lines, the program then searches to whatever time *
 *   or depth limit has been set, when it reaches the end-of-file condition or *
 *   when it reads a record containing the string "end" it then displays the   *
 *   number correct and the number missed.                                     *
 *                                                                             *
 *******************************************************************************
 */
void Test(char *filename)
{
  FILE *test_input;
  int i, move, right = 0, wrong = 0, correct;
  int time = 0, time_solved = 0, len;
  UINT64 nodes = 0, evaluations = 0, quiescent_nodes = 0;
  UINT64 nodes_ps = 0;
  char *eof;
  float sum_depth = 0.0;
  float score_avg = 0.0;
  int depth;
  TREE *const tree = block[0];
  int value, score;
  int scores = 0;
  float sum_score = 0.0;
  float evaluations_p = 0.0;
  float quiescent_nodes_p = 0.0;
  int mates = 0;
  char file_name[1024];

/*
 ************************************************************
 *                                                          *
 *   read in the position and then the solutions.  after    *
 *   executing a search to find the best move (according    *
 *   to the program, anyway) compare it against the list    *
 *   of solutions and count it right or wrong.              *
 *                                                          *
 ************************************************************
 */
  if (!(test_input = fopen(filename, "r"))) {
    printf("file %s does not exist.\n", filename);
    return;
  }
  Print(4095, "\n");
  eof = fgets(buffer, 4096, test_input);
  if (strstr(buffer, "title")) ;
  else {
    fclose(test_input);
    TestEPD(filename);
    return;
  }
  strcpy(file_name, filename);
  if (book_file) {
    fclose(book_file);
    book_file = 0;
  }
  if (books_file) {
    fclose(books_file);
    books_file = 0;
  }
  rewind(test_input);
  test_score_diff = 0;
  test_running = 1;
  while (1) {
    eof = fgets(buffer, 4096, test_input);
    if (eof) {
      char *delim;

      delim = strchr(buffer, '\n');
      if (delim)
        *delim = 0;
      delim = strchr(buffer, '\r');
      if (delim)
        *delim = ' ';
    } else
      break;
    nargs = ReadParse(buffer, args, " ;");
    if (!strcmp(args[0], "end"))
      break;
    else if (!strcmp(args[0], "title")) {
      Print(4095,
          "=============================================="
          "=================================\n");
      Print(4095, "! ");
      len = 0;
      for (i = 1; i < nargs; i++) {
        Print(4095, "%s ", args[i]);
        len += strlen(args[i]) + 1;
        if (len > 65)
          break;
      }
      for (i = len; i < 67; i++)
        printf(" ");
      Print(4095, "         !\n");
      Print(4095,
          "=============================================="
          "=================================\n");
    } else if (strcmp(args[0], "solution")) {
      Option(tree);
    } else {
      number_of_solutions = 0;
      solution_type = 0;
      Print(4095, "solution ");
      for (i = 1; i < nargs; i++) {
        if (args[i][strlen(args[i]) - 1] == '?') {
          solution_type = 1;
          args[i][strlen(args[i]) - 1] = '\0';
        } else if (*(args + i)[strlen(args[i]) - 1] == '!') {
          solution_type = 0;
          args[i][strlen(args[i]) - 1] = '\0';
        }
        move = InputMove(tree, args[i], 0, wtm, 0, 0);
        if (move) {
          solutions[number_of_solutions] = move;
          /*Print(4095, "%d. %s", (number_of_solutions++) + 1, OutputMove(tree,
                  move, 0, wtm));*/
          Print(4095, "%s", OutputMove(tree, move, 0, wtm));
          number_of_solutions++;
          if (solution_type == 1)
            Print(4095, "? ");
          else
            Print(4095, "  ");
        } else
          DisplayChessBoard(stdout, tree->pos);
      }
      Print(4095, "\n");
      InitializeHashTables();
      last_pv.pathd = 0;
      thinking = 1;
      tree->position[1] = tree->position[0];
      value = Iterate(wtm, think, 0);
      score = value;
      thinking = 0;
      nodes += tree->nodes_searched;
      quiescent_nodes += tree->quiescent_nodes_searched;
      evaluations += tree->evaluations;
      depth = iteration_depth;
      sum_depth += (float) depth;
      if (Abs(score) <= NOT_MATE) {
        sum_score += Abs(score);
        scores++;
      }
      else {
        mates++;
      }
      time += (end_time - start_time);
      correct = solution_type;
      for (i = 0; i < number_of_solutions; i++) {
        if (!solution_type) {
          if (solutions[i] == tree->pv[1].path[1])
            correct = 1;
        } else if (solutions[i] == tree->pv[1].path[1])
          correct = 0;
      }
      if (correct) {
        if (Abs(score) <= NOT_MATE) {
          if (score < test_score_threshold) {
            correct = 0;
          }
        }
      }
      if (correct)
        right++;
      else
        wrong++;
      if (correct) {
        time_solved += (end_time - start_time);
      }
      if (correct) {
        Print(4095, "----------------------> solution correct (%4d/%4d) %s\n",
            right, right + wrong, DisplayTime(time));
      }
      else {
        Print(4095, "----------------------> solution incorrect (%4d/%4d) %s\n",
            right, right + wrong, DisplayTime(time));
      }
    }
  }
  if (time)
    nodes_ps = nodes * 100 / time;
  if (nodes > 0)
    quiescent_nodes_p = (float)quiescent_nodes * 100 / (float)nodes;
  if (nodes > 0)
    evaluations_p = (float)evaluations * 100 / (float)nodes;
  if (scores > 0)
    score_avg = (float)sum_score / (float)scores;
/*
 ************************************************************
 *                                                          *
 *   now print the results.                                 *
 *                                                          *
 ************************************************************
 */
  if (right + wrong) {
    Print(4095, "\n\n\n");
    Print(4095, "-----------------------------------------------------------\n");
    Print(4095, "Test Results Summary for %s:\n", file_name);
    Print(4095, "total positions searched.......................%12d\n", right + wrong);
    Print(4095, "number right...................................%12d\n", right);
    Print(4095, "number wrong...................................%12d\n", wrong);
    Print(4095, "percentage right...............................%12.1f\n",
        (float)right * 100 / (float)(right + wrong));
    Print(4095, "percentage wrong...............................%12.1f\n",
        (float)wrong * 100 / (float)(right + wrong));
    Print(4095, "test score threshold...........................%+12d\n", test_score_threshold);
    Print(4095, "test search time limit.........................%12.0f\n", (float)search_time_limit / 100.);
    Print(4095, "total nodes searched...........................%12llu\n", nodes);
    Print(4095, "nodes evaluated................................%12llu\n", evaluations);
    Print(4095, "quiescent nodes searched*......................%12llu\n", quiescent_nodes);
    Print(4095, "nodes evaluated percentage.....................%12.1f\n", evaluations_p);
    Print(4095, "quiescent nodes percentage.....................%12.1f\n", quiescent_nodes_p);
    Print(4095, "average search depth...........................%12.3f\n",
        (float)sum_depth / (float)(right + wrong));
    Print(4095, "average nodes per second.......................%12llu\n", nodes_ps);
    Print(4095, "total time.....................................%12s\n", DisplayTime(time));
    if (right) {
      Print(4095, "total time solved..............................%12s\n", DisplayTime(time_solved));
      Print(4095, "average time solved............................%12s\n", DisplayTime(time_solved / right));
    }
    Print(4095, "mates found....................................%12d\n", mates);
    Print(4095, "average absolute score**.......................%12.1f\n", score_avg);
    Print(4095, "-----------------------------------------------------------\n");
    Print(4095, "* nodes searched in quiescent search\n");
    Print(4095, "** not counting mate scores\n");
  }
  input_stream = stdin;
  test_running = 0;
  early_exit = 99;
  test_score_threshold = 0;
}

/* last modified 05/31/19 */
/*
 *******************************************************************************
 *                                                                             *
 *   TestEPD() is used to test the program against a suite of test positions to*
 *   measure its performance on a particular machine, or to evaluate its skill *
 *   after modifying it in some way.                                           *
 *                                                                             *
 *   the test is initiated by using the "test <filename>" command to read in   *
 *   the suite of problems from file <filename>.  the format of this file is   *
 *   as follows:                                                               *
 *                                                                             *
 *   <forsythe-string>  am/bm move1 move2 etc; title "xxx"                     *
 *                                                                             *
 *   am means "avoid move" and bm means "best move".  each test position may   *
 *   have multiple moves to avoid or that are best, but both am and bm may not *
 *   appear on one position.                                                   *
 *                                                                             *
 *   the title is just a comment that is given in the program output to make it*
 *   easier to match output to specific positions.                             *
 *                                                                             *
 *******************************************************************************
 */
void TestEPD(char *filename)
{
  FILE *test_input;
  int i, move, right = 0, wrong = 0, correct;
  int time = 0, len;
  UINT64 nodes = 0;
  char *eof, *mvs, *title;
  float avg_depth = 0.0;
  TREE *const tree = block[0];

/*
 ************************************************************
 *                                                          *
 *   read in the position and then the solutions.  after    *
 *   executing a search to find the best move (according    *
 *   to the program, anyway) compare it against the list    *
 *   of solutions and count it right or wrong.              *
 *                                                          *
 ************************************************************
 */
  if (!(test_input = fopen(filename, "r"))) {
    printf("file %s does not exist.\n", filename);
    return;
  }
  if (book_file) {
    fclose(book_file);
    book_file = 0;
  }
  if (books_file) {
    fclose(books_file);
    books_file = 0;
  }
  while (1) {
    eof = fgets(buffer, 4096, test_input);
    if (eof) {
      char *delim;

      delim = strchr(buffer, '\n');
      if (delim)
        *delim = 0;
      delim = strchr(buffer, '\r');
      if (delim)
        *delim = ' ';
    } else
      break;
    mvs = strstr(buffer, " bm ");
    if (!mvs)
      mvs = strstr(buffer, " am ");
    if (!mvs) {
      Print(4095, "Error am/bm field missing, input string follows\n%s\n",
          buffer);
      continue;
    }
    mvs++;
    title = strstr(buffer, "id");
    *(mvs - 1) = 0;
    if (title)
      *(title - 1) = 0;
    if (title) {
      title = strchr(title, '\"') + 1;
      if (title) {
        if (strchr(title, '\"')) {
          *strchr(title, '\"') = 0;
        }
      }
      Print(4095,
          "=============================================="
          "========================\n");
      Print(4095, "! ");
      Print(4095, "%s ", title);
      len = 66 - strlen(title);
      for (i = 0; i < len; i++)
        printf(" ");
      Print(4095, "!\n");
      Print(4095,
          "=============================================="
          "========================\n");
    }
    Option(tree);
    nargs = ReadParse(mvs, args, " ;");
    number_of_solutions = 0;
    solution_type = 0;
    if (!strcmp(args[0], "am"))
      solution_type = 1;
    Print(4095, "solution ");
    for (i = 1; i < nargs; i++) {
      move = InputMove(tree, args[i], 0, wtm, 0, 0);
      if (move) {
        solutions[number_of_solutions] = move;
        Print(4095, "%d. %s", (number_of_solutions++) + 1, OutputMove(tree,
                move, 0, wtm));
        if (solution_type == 1)
          Print(4095, "? ");
        else
          Print(4095, "  ");
      } else
        DisplayChessBoard(stdout, tree->pos);
    }
    Print(4095, "\n");
    InitializeHashTables();
    last_pv.pathd = 0;
    thinking = 1;
    tree->position[1] = tree->position[0];
    (void) Iterate(wtm, think, 0);
    thinking = 0;
    nodes += tree->nodes_searched;
    avg_depth += (float) iteration_depth;
    time += (end_time - start_time);
    correct = solution_type;
    for (i = 0; i < number_of_solutions; i++) {
      if (!solution_type) {
        if (solutions[i] == tree->pv[1].path[1])
          correct = 1;
      } else if (solutions[i] == tree->pv[1].path[1])
        correct = 0;
    }
    if (correct) {
      right++;
      Print(4095, "----------------------> solution correct (%d/%d).\n", right,
          right + wrong);
    } else {
      wrong++;
      Print(4095, "----------------------> solution incorrect (%d/%d).\n",
          right, right + wrong);
    }
  }
/*
 ************************************************************
 *                                                          *
 *   now print the results.                                 *
 *                                                          *
 ************************************************************
 */
  if (right + wrong) {
    Print(4095, "\n\n\n");
    Print(4095, "test results summary:\n\n");
    Print(4095, "total positions searched..........%12d\n", right + wrong);
    Print(4095, "number right......................%12d\n", right);
    Print(4095, "number wrong......................%12d\n", wrong);
    Print(4095, "percentage right..................%12.1f\n",
        (float)right * 100 / (float)(right + wrong));
    Print(4095, "percentage wrong..................%12.1f\n",
        (float)wrong * 100 / (float)(right + wrong));
    Print(4095, "total nodes searched..............%12llu\n", nodes);
    Print(4095, "average search depth..............%12.1f\n",
        (float)avg_depth / (float)(right + wrong));
    Print(4095, "nodes per second..................%12llu\n", nodes * 100 / time);
    Print(4095, "total time........................%12s\n", DisplayTime(time));
  }
  input_stream = stdin;
  early_exit = 99;
}

